from collections import deque
from enum import Enum
from multiprocessing.connection import Client
import queue
from threading import Thread
import traceback
from torch.distributed.rpc import RRef, rpc_async, remote
from Sample.neighbor_sampler import NeighborSampler
import time
import torch.distributed as dist
import torch.multiprocessing as mp
import graph_store
import os
import time
from message_worker import _check_future_finish, _get_batch_data, _sample_node_neighbors_server
from rpc_server import  RpcCommand, close_rpc,  start_rpc_listener,start_rpc_caller
import distparser as parser
import sys
WORKER_RANK = parser._get_worker_rank()
NUM_SAMPLER = parser._get_num_sampler()
WORLD_SIZE = parser._get_world_size()
QUEUE_SIZE =parser._get_queue_size()
MAX_QUEUE_SIZE = parser._get_max_queue_size()
RPC_NAME = parser._get_RPC_NAME()

class MpCommand(Enum):
    """Enum class for multiprocessing command"""

    INIT_RPC = 0  # Not used in the task queue
    SET_COLLATE_FN = 1
    CALL_BARRIER = 2
    DELETE_COLLATE_FN = 3
    CALL_COLLATE_FN = 4
    CALL_FN_ALL_WORKERS = 5
    FINALIZE_POOL = 6

start = time.time()
def wait_thread(waitfut,data_queue):
    global start
    while(True):
        if len(waitfut) > 0:
            result = waitfut[0]
            future_to_check = result.get("union_args")[2]
            #print('check',time.time()-start)
            if(_check_future_finish(future_to_check) is True):
                print('finish',start)
                print('wait for future',time.time()-start)
                sys.stdout.flush()
                start = time.time()
                batch_data = _get_batch_data(result)
                waitfut.popleft()
                data_queue.put(
                (
                    result.get("data_name"),
                    batch_data,
                )
                    
            )
def init_process(sampler_id,rpc_config,comm_config):
    """start the proxy of rpc on the process """
    rpc_master_addr,rpc_port,num_worker_threads = rpc_config
    start_rpc_caller(rpc_master_addr,
                    rpc_port,
                    {
                        "num_worker_threads": num_worker_threads,
                        "rpc_name": RPC_NAME,
                        "rpc_world_size": WORLD_SIZE * (NUM_SAMPLER + 1),
                        "worker_rank":  WORKER_RANK,
                        "rpc_worker_rank": (NUM_SAMPLER + 1) * WORKER_RANK + sampler_id + 1}
                    )
    print('start work')
    global start
    try:
        data_queue,task_queue,barrier = comm_config
        collate_fn_dict = {}
        sampler_dict = {}
        keep_polling = True
        waitfut = deque()
        thread = Thread(target=wait_thread,args=(waitfut,data_queue))
        thread.start()
        while keep_polling or graph_store._get_size() > 0:
            if(len(waitfut)>QUEUE_SIZE):
                continue
            try:
                command,args = task_queue.get(timeout=5)
            except queue.Empty:
                continue
            if command == MpCommand.SET_COLLATE_FN:
                dataloader_name,graph_inshm,func,sampler_info = args
                graph_store._set_graph(dataloader_name,graph_inshm)
                collate_fn_dict[dataloader_name] = func
                #neighbors,deg = graph_inshm._get_sampler_from_shame()
                sampler = sampler_info #NeighborSampler(*sampler_info,neighbors= neighbors,deg = deg)
                sampler_dict[dataloader_name] = sampler

            elif command == MpCommand.CALL_BARRIER:
                barrier.wait()

            elif command == MpCommand.DELETE_COLLATE_FN:
                (dataloader_name,) = args
                del collate_fn_dict[dataloader_name]
                del sampler_dict[dataloader_name]
                graph_inshm = graph_store._get_graph(dataloader_name)
                graph_store._del_graph(dataloader_name)
                graph_inshm._close_graph_in_shame()
                if(barrier.wait()==0):
                    graph_inshm._unlink_graph_in_shame()

            elif command == MpCommand.CALL_COLLATE_FN:
                dataloader_name, collate_args = args
                neighbor_nodes, sampled_edge_index= sampler_dict[dataloader_name].sample_from_nodes(collate_args)
                union_args=_sample_node_neighbors_server(dataloader_name,neighbor_nodes)
                print('sample time',time.time()-start)
                start = time.time()
                waitfut.append({
                    "data_name":dataloader_name,
                    "input_data":collate_args,
                    "nids":neighbor_nodes,
                    "union_args":union_args,
                    "edge_index":sampled_edge_index}
                    )
            
                #data_queue.put(
                #    (
                #        dataloader_name,
                #        collate_fn_dict[dataloader_name](collate_args),
                #    )
                #)
            elif command == MpCommand.CALL_FN_ALL_WORKERS:
                func, func_args = args
                func(func_args)
            elif command == MpCommand.FINALIZE_POOL:
                keep_polling = False 
                graph_store._clear_all(barrier)
                close_rpc()
            else:
                raise Exception("Unknown command")
            
    except Exception as e:
        traceback.print_exc()
        raise e
    
custom_pool = None
rpc_server_prco = None
rpc_server_queue = None
rpc_server_barrier = None
keep_polling = True
class CustomPool:
    def __init__(self,rpc_config):
        self.max_queue_size = MAX_QUEUE_SIZE
        self.num_samplers = NUM_SAMPLER
        ctx = mp.get_context("spawn")
        self.result_queue = ctx.Queue(self.max_queue_size)
        self.results = {}
        self.task_queues = []
        self.process_list = []
        self.current_proc_id = 0
        self.cache_result_dict = {}
        self.barrier = ctx.Barrier(self.num_samplers)
        for rank in range(self.num_samplers):
            task_queue = ctx.Queue(self.max_queue_size)
            self.task_queues.append(task_queue)
            proc = ctx.Process(
                target=init_process,
                args=(
                    rank,rpc_config,
                    (self.result_queue,task_queue,self.barrier),
                )
                    
            )
            proc.daemon=True
            proc.start()
            self.process_list.append(proc)
        self.call_barrier()

        
    def set_collate_fn(self,func,sampler,dataloader_name,graph_inshm):
        for i in range(self.num_samplers):
            self.task_queues[i].put(
                (MpCommand.SET_COLLATE_FN,(dataloader_name,graph_inshm,func,sampler))
            )
        global rpc_server_queue
        rpc_server_queue.put((RpcCommand.SET_GRAPH,(dataloader_name,graph_inshm)))
        rpc_server_queue.put((RpcCommand.CALL_BARRIER,tuple()))
        global rpc_server_barrier
        rpc_server_barrier.wait()
        self.call_barrier()
        self.results[dataloader_name]=[]

    def submit_task(self, dataloader_name, args):
        """Submit task to workers"""
        # Round robin
        self.task_queues[self.current_proc_id].put(
            (MpCommand.CALL_COLLATE_FN, (dataloader_name, args))
        )
        self.current_proc_id = (self.current_proc_id + 1) % self.num_samplers
        
    def submit_task_to_all_workers(self, func, args):
        """Submit task to all workers"""
        for i in range(self.num_samplers):
            self.task_queues[i].put(
                (MpCommand.CALL_FN_ALL_WORKERS, (func, args))
            )
    def get_result(self, dataloader_name, timeout=1800):
        """Get result from result queue"""
        if dataloader_name not in self.results:
            raise Exception(
                f"Got result from an unknown dataloader {dataloader_name}."
            )
        while len(self.results[dataloader_name]) == 0:
            data_name, data = self.result_queue.get(timeout=timeout)
            self.results[data_name].append(data)
        return self.results[dataloader_name].pop(0)

    def delete_collate_fn(self, dataloader_name):
        """Delete collate function"""
        print('del dataloader')
        global rpc_server_queue
        rpc_server_queue.put(
            (RpcCommand.UNLOAD_GRAPH, (dataloader_name,))
        )
        global rpc_server_barrier
        if(keep_polling is True):
            rpc_server_barrier.wait()
        print('del dataloader')
        for i in range(self.num_samplers):
            self.task_queues[i].put(
                (MpCommand.DELETE_COLLATE_FN, (dataloader_name,))
            )
        if dataloader_name in self.results:
            del self.results[dataloader_name] 
    
    def call_barrier(self):
        """Call barrier at all workers"""
        for i in range(self.num_samplers):
            self.task_queues[i].put((MpCommand.CALL_BARRIER, tuple()))

    def close(self):
        """Close worker pool"""
        for i in range(self.num_samplers):
            self.task_queues[i].put(
                (MpCommand.FINALIZE_POOL, tuple()), block=False
            )
        time.sleep(0.5)  # Fix for early python version

    def join(self):
        """Join the close process of worker pool"""
        for i in range(self.num_samplers):
            self.process_list[i].join()
    



def get_sampler_pool():
    global custom_pool
    return custom_pool
def init_distribution(master_addr=None,master_port=None,rpc_master_addr=None,rpc_port=None, num_worker_threads = 16):
    print('init distribution')
    if master_addr is None or master_port is None:
        raise Exception(
                f"The master address is unknown."
            )
    if rpc_port is None:
        raise Exception(
                f"The rpc listener address is unknown."
            )
    init_method="tcp://{}:{}".format(master_addr,master_port)
    global custom_pool
    custom_pool = CustomPool((rpc_master_addr,rpc_port,num_worker_threads))
    world_size = parser._get_world_size()
    worker_rank = parser._get_worker_rank()
    num_sampler = parser._get_num_sampler()
    print("!",world_size,worker_rank)
    dist.init_process_group(backend="gloo", world_size=world_size, rank=worker_rank,init_method=init_method)
    ctx = mp.get_context("spawn")
    global rpc_server_queue
    rpc_server_queue = ctx.Queue(MAX_QUEUE_SIZE)
    global rpc_server_barrier
    rpc_server_barrier = ctx.Barrier(2)
    rpc_server_proc= ctx.Process(target = start_rpc_listener,
                                 args=(rpc_master_addr,
                                       rpc_port,
                                       {"num_worker_threads": num_worker_threads,
                                        "rpc_name": RPC_NAME,
                                        "rpc_world_size": world_size * (num_sampler + 1),
                                        "worker_rank":  worker_rank,
                                        "rpc_worker_rank": (num_sampler + 1) * worker_rank
                                        },
                                        (rpc_server_queue,rpc_server_barrier)),
                                 )
    rpc_server_proc.start()
    dist.barrier()
    
def close_distribution():
    dist.barrier()
    print('del')
    global rpc_server_queue
    rpc_server_queue.put((RpcCommand.STOP_RPC,tuple()))
    global rpc_server_barrier
    rpc_server_barrier.wait()
    global custom_pool
    custom_pool.close()
    global keep_polling 
    keep_polling = False
    #rpc.init_rpc(OBSERVER_NAME.format(rank),rank=rank,world_size=world_size)
    #if(rank==0):
    #    message_worker.init_sampler_worker()
    #dist.barrier()
    #print(OBSERVER_NAME.format(rank))
    #worker = sampler_worker(world_size,OBSERVER_NAME)
    #if(rank==0):
    #rrefs = 
    #master_worker = sampler_worker(world_size,OBSERVER_NAME)
