from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from multiprocessing import Barrier
from multiprocessing.connection import Client
import queue
import sys
import argparse
import traceback
from torch.distributed.rpc import RRef, rpc_async, remote
import torch.distributed.rpc as rpc
import torch
from part.Utils import GraphData
from Sample.neighbor_sampler import NeighborSampler
import time
from typing import Optional
import torch.distributed as dist
import torch.multiprocessing as map
from BatchData import BatchData
import asyncio 
import os
from Sample.neighbor_sampler import NeighborSampler
from graph_store import _get_graph
from shared_graph_memory import GraphInfoInShm
import distparser as parser

WORKER_RANK = parser._get_worker_rank()
NUM_SAMPLER = parser._get_num_sampler()
WORLD_SIZE = parser._get_world_size()
QUEUE_SIZE =parser._get_queue_size()
MAX_QUEUE_SIZE = parser._get_max_queue_size()
RPC_NAME = parser._get_RPC_NAME()



#@rpc.functions.async_execution
def _get_local_attr(data_name,nodes):   
    print('get_local_attr',time.time()) 
    graph = _get_graph(data_name)
    local_id = graph.get_localId_by_partitionId(parser._get_worker_rank(),nodes)
    return graph.select_attr(local_id)

def _request_remote_attr(rank,data_name,nodes):
    print('request remote attr ',rank,data_name)
    fut = rpc_async(
        parser._get_RPC_NAME().format(rank),
        _get_local_attr,
        args=(data_name,nodes,)
    )
    return fut

def _request_all_remote_attr(data_name,nodes_list):
    worker_size = parser._get_world_size()
    worker_rank = parser._get_worker_rank()
    futs = []
    for rank in range(worker_size):
        if(rank == worker_rank):
            futs.append(None)
            continue
        else:
            if(nodes_list[rank].size(0) == 0):
                futs.append(None)
            else:
                futs.append(_request_remote_attr(rank,data_name,nodes_list[rank]))
    return futs

def _check_future_finish(futs):
    check = True
    for fut in futs:
        if fut is not None and fut.done() is False:
            check = False
    return check

def _split_node_part(data_name,node_id_list):
    graph = _get_graph(data_name)
    part_list=[]
    scatter_list=[]
    worker_size = parser._get_world_size()
    for rank in range(worker_size):
        group_id=(graph.partptr[rank]<=node_id_list) & (node_id_list<graph.partptr[rank+1])
        part_list.append(torch.masked_select(node_id_list,group_id))
        scatter_list.append(torch.nonzero(group_id)) 
    scatter_index=torch.cat(scatter_list)
    print('part_list',part_list)
    return part_list,scatter_index
    
def _union_remote_and_local(data_name,scatter_list,local_nodes,futs):
    worker_rank = parser._get_worker_rank()
    worker_size = parser._get_world_size()
    feature_part = []
    local_feature=_get_local_attr(data_name,local_nodes)
    node_feature_list = []
    for rank in range(worker_size):
        if(rank == worker_rank):
            node_feature_list.append(local_feature)
        elif(futs[rank] is None):
            continue
        else:
            node_feature_list.append(futs[rank].value())

    node_feature = torch.cat(node_feature_list)
    attr_size=node_feature.size()
    x=torch.zeros(attr_size)
    x=x.scatter(0,scatter_list.repeat(1,attr_size[1]),node_feature)
    return x

def _get_batch_data(kwargs):
    data_name = kwargs.get("data_name")
    input_data = kwargs.get("input_data")
    nids = kwargs.get("nids")
    scatter_list,local_nodes,futs = kwargs.get("union_args")
    edge_index = kwargs.get("edge_index")
    x=_union_remote_and_local(data_name,scatter_list,local_nodes,futs)
    graph = _get_graph(data_name)
    worker_rank = parser._get_worker_rank()
    local_y_id = graph.get_localId_by_partitionId(worker_rank,input_data)
    y = graph.select_y(local_y_id)
    #print(t2-t1,t3-t2,t4-t3,t5-t4,time.time()-t4)
    return BatchData(nids,edge_index,roots=input_data,x=x,y=y)

def _sample_node_neighbors_server(data_name,neighbor_nodes):
    '''
    sample the struct of the subgraph
    '''
    print('sample node neighbors server')
    worker_rank = parser._get_worker_rank()
    part_feature_list,scatter_list = _split_node_part(data_name,neighbor_nodes)
    futs = _request_all_remote_attr(data_name,part_feature_list)
    return scatter_list,part_feature_list[worker_rank],futs

